////////////////////////////////////////////////////////////////////////
//
// glue.cpp
// Cellular Automata Machine Engine Glue for Python and SWIG
//
// Copyright (C) 1985-2000 by Don Hopkins
// All rights reserved.
//
////////////////////////////////////////////////////////////////////////


#include <stdio.h>
#include "cell.h"
#include "glue.h"
#include <Python.h>


////////////////////////////////////////////////////////////////////////


extern Pycairo_CAPI_t *Pycairo_CAPI;


////////////////////////////////////////////////////////////////////////


Cam *MakeCam()
{
    return
        new Cam();
}


void DestroyCam(Cam *cam)
{
    if (cam != NULL) {
        delete cam;
    } // if
}


PyObject *MakeBuffer(
    int width,
    int height)
{
    int size = width * height;
    PyObject *result =
        PyBuffer_New(size);

    unsigned char *buf = NULL;
    int len = 0;
    int n =
        PyObject_AsWriteBuffer(
            result,
            (void **)&buf,
            &len);
    int i;
    for (i = 0; i < size; i++) {
        buf[i] = rand() >> 3;
    } // for i

    return result;
}


PyObject *MakePalette()
{
    return MakeBuffer(256 * 3, 1);
}


static unsigned long GetCell(
    int row,
    int col,
    void *bufData,
    int colBytes,
    int rowBytes,
    char typeCode)
{
  switch (typeCode) {

  case 'c':
  case 'b':
  case 'B':
    return
      (unsigned long)(
	*(unsigned char *)(
	  (unsigned char *)bufData +
	  (col * colBytes) +
	  (row * rowBytes)));
    break;

  case 'u':
  case 'h':
  case 'H':
  case 'i':
  case 'I':
    return
      (unsigned long)(
	*(unsigned short *)(
	  (unsigned char *)bufData +
	  (col * colBytes) +
	  (row * rowBytes)));

  case 'l':
  case 'L':
    return
      (unsigned long)(
	*(unsigned long *)(
	  (unsigned char *)bufData +
	  (col * colBytes) +
	  (row * rowBytes)));

  case 'f':
    return
      (unsigned long)(
	*(float *)(
	  (unsigned char *)bufData +
	  (col * colBytes) +
	  (row * rowBytes)));

  case 'd':
    return
      (unsigned long)(
	*(double *)(
	  (unsigned char *)bufData +
	  (col * colBytes) +
	  (row * rowBytes)));

  default:
    return 0;

  }
}


void RenderCellTiles(
    cairo_t *ctx,
    PyObject *tiles,
    int tilesWidth,
    int tilesHeight,
    PyObject *tileMap,
    int tileSize,
    PyObject *buf,
    int width,
    int height,
    int colBytes,
    int rowBytes,
    char typeCode,
    int renderCol,
    int renderRow,
    int renderCols,
    int renderRows,
    double alpha)
{
  int r;
  int c;
  
  if (!PyObject_TypeCheck(
	tiles,
	&PycairoSurface_Type)) {
    PyErr_SetString(
      PyExc_TypeError,
      "expected cairo.Surface");
    return;
  }

  cairo_surface_t *tilesSurf =
    PycairoSurface_GET(
      tiles);

  unsigned int tileCount =
    PySequence_Size(tileMap);

  const int *tileMapData = NULL;
  Py_ssize_t tileMapLength = 0;

  if (PyObject_AsReadBuffer(
        tileMap,
        (const void **)&tileMapData,
        &tileMapLength) != 0) {
    PyErr_SetString(
      PyExc_TypeError,
      "expected array");
    return;
  }

  if ((tileMapLength / sizeof(unsigned int)) != tileCount) {
    PyErr_SetString(
      PyExc_TypeError,
      "expected array of 4 byte integers");
    return;
  }

  void *bufData = NULL;
  Py_ssize_t bufLength = 0;

  if (PyObject_AsReadBuffer(
        buf,
        (const void **)&bufData,
        &bufLength) != 0) {
    PyErr_SetString(
      PyExc_TypeError,
      "expected buffer");
    return;
  }

  if ((height * rowBytes * colBytes) > bufLength) {
    PyErr_SetString(
      PyExc_TypeError,
      "buffer not big enough");
    return;
  }

  int tilesPerRow = 
    tilesWidth / tileSize;

  int renderX =
    renderCol * tileSize;
  int renderY =
    renderRow * tileSize;

  for (r = 0; r < renderRows; r++) {

    for (c = 0; c  < renderCols; c++) {

      int col = (renderCol + c) % width;
      int row = (renderRow + r) % height;

      unsigned long cell = 
	GetCell(
	  row,
	  col,
	  bufData,
	  colBytes,
	  rowBytes,
	  typeCode);

      int tileIndex =
	cell % tileCount;

      int tile =
	tileMapData[tileIndex];

      double x = 
	tileSize * col;
      double y =
	tileSize * row;

      int tileCol = 
	tile % tilesPerRow;
      int tileRow =
	tile / tilesPerRow;

      cairo_save(
	ctx);

      int tileX =
	tileCol * tileSize;
      int tileY =
	tileRow * tileSize;

      cairo_translate(
	ctx,
	x - renderX,
	y - renderY);

      cairo_rectangle(
        ctx,
	0,
	0,
	tileSize,
	tileSize);

      cairo_clip(
        ctx);

      cairo_set_source_surface(
	ctx,
	tilesSurf,
	-tileX,
	-tileY);

      if (alpha >= 1.0) {
	cairo_paint(
	  ctx);
      } else {
	cairo_paint_with_alpha(
	  ctx,
	  alpha);
      }

      cairo_restore(
	ctx);

    }
  }
}


void RenderCellPixels(
    cairo_surface_t *destSurf,
    cairo_surface_t *cmapSurf,
    PyObject *tileMap,
    PyObject *buf,
    int width,
    int height,
    int colBytes,
    int rowBytes,
    char typeCode,
    int renderCol,
    int renderRow,
    int renderCols,
    int renderRows)
{
  int r;
  int c;

  unsigned int tileCount =
    PySequence_Size(tileMap);

  const int *tileMapData = NULL;
  Py_ssize_t tileMapLength = 0;

  if (PyObject_AsReadBuffer(
        tileMap,
        (const void **)&tileMapData,
        &tileMapLength) != 0) {
    PyErr_SetString(
      PyExc_TypeError,
      "expected array");
    return;
  }

  if ((tileMapLength / sizeof(unsigned int)) != tileCount) {
    PyErr_SetString(
      PyExc_TypeError,
      "expected array of 4 byte integers");
    return;
  }

  void *bufData = NULL;
  Py_ssize_t bufLength = 0;

  if (PyObject_AsReadBuffer(
        buf,
        (const void **)&bufData,
        &bufLength) != 0) {
    PyErr_SetString(
      PyExc_TypeError,
      "expected buffer");
    return;
  }

  if ((height * rowBytes * colBytes) > bufLength) {
    PyErr_SetString(
      PyExc_TypeError,
      "buffer not big enough");
    return;
  }

  unsigned char *destData =
    cairo_image_surface_get_data(destSurf);
  cairo_format_t destFormat =
    cairo_image_surface_get_format(destSurf);
  int destWidth =
    cairo_image_surface_get_width(destSurf);
  int destHeight =
    cairo_image_surface_get_height(destSurf);
  int destStride =
    cairo_image_surface_get_stride(destSurf);

  //printf("destData %x destFormat %d destWidth %d destHeight %d destStride %d\n",
  //	 destData, destFormat, destWidth, destHeight, destStride);

  unsigned char *cmapData =
    cairo_image_surface_get_data(cmapSurf);
  cairo_format_t cmapFormat =
    cairo_image_surface_get_format(cmapSurf);
  int cmapWidth =
    cairo_image_surface_get_width(cmapSurf);
  int cmapHeight =
    cairo_image_surface_get_height(cmapSurf);
  int cmapStride =
    cairo_image_surface_get_stride(cmapSurf);

  //printf("cmapData %x cmapFormat %d cmapWidth %d cmapHeight %d cmapStride %d\n",
  //	 cmapData, cmapFormat, cmapWidth, cmapHeight, cmapStride);

  //printf("========================================================================\n");

  for (r = 0; r < renderRows; r++) {

    for (c = 0; c  < renderCols; c++) {

      int col = (renderCol + c) % width;
      int row = (renderRow + r) % height;

      unsigned long cell = 
	GetCell(
	  row,
	  col,
	  bufData,
	  colBytes,
	  rowBytes,
	  typeCode);

      int tileIndex =
	cell % tileCount;

      int tile =
	tileMapData[tileIndex];

      int sourceX = 
	tile % cmapWidth;
      int sourceY =
	tile / cmapWidth;
      unsigned char *sourcePixel =
	cmapData +
	(sourceX * 4) +
	(sourceY * cmapStride);
      
      unsigned char *destPixel =
	destData +
	(c * 4) +
	(r * destStride);

      *(long *)destPixel = *(long *)sourcePixel;

#if 0
      printf("r %d c %d cell %d sourceX %d sourceY %d sourcePixel %x %08x destPixel %x %08x\n",
	     r,
	     c,
	     cell,
	     sourceX,
	     sourceY,
	     sourcePixel,
	     *(long *)sourcePixel,
	     destPixel,
	     *(long *)destPixel);
#endif

    }
  }
}


////////////////////////////////////////////////////////////////////////
